# Copyright 2005 PyTom <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import os.path
import os
import re
import zipfile
import cStringIO

import alice

class Doll(object):

    def __init__(self, name):
        self.name = name

        # A map from category name to map from layer name to layer.
        self.layers = { }

        # The configuration for this object.
        self.config = { "credits_and_copyrights" : [ ] }

    # Returns the size of the base image.
    def sizes(self):
        
        # Find an image with 0 priority.
        for cat in self.layers.itervalues():
            for layer in cat.itervalues():
                for image in layer.images:
                    if image.priority == 0:
                        return alice.renderer.load_spng(image)[0]

        return (100, 100)
        

class Layer(object):

    def __init__(self, doll, category, name):

        # A reference to the doll containing this layer.
        self.doll = doll

        # The name of the category this layer is in.
        self.category = category

        # The name of this layer.
        self.name = name

        # A list of image objects in this layer.
        self.images = [ ]

        # The path to the thumbnail of this layer.
        self.thumb = None

class Image(object):

    def __init__(self, layer, name, priority, path):

        # A reference to the layer object containing this image.
        self.layer = layer

        # The name of this image.
        self.name = name

        # The priority of this image. A floating point number, with
        # smaller numbers being closer to the user, and larger numbers
        # being further away.
        self.priority = priority

        # The path to the image, as a Path object given below.
        self.path = path

        
# A map from doll name to the Doll object representing that doll.
dolls = { }

class Path(object):
    """
    This is used to represent a doll component stored in a file. Eventually,
    a duck-type-compatible version of this may be used to find doll components
    stored in other formats.
    """

    def __init__(self, filename):
        self.filename = filename
        self.basename = os.path.basename(filename)

    def __repr__(self):
        return "<path %r>" % self.filename

    def read(self):
        return file(self.filename, 'rb').read()

    def open(self):
        return file(self.filename, 'rb')

class ZipPath(object):
    """
    This is a Path to a component found in a zipfile.
    """

    def __init__(self, zipfile, filename):
        self.zipfile = zipfile
        self.filename = filename
        self.basename = os.path.basename(filename)

    def __repr__(self):
        return "<path %r %r>" % (self.zipfile, self.filename)

    def read(self):
        return self.zipfile.read(self.filename)

    def open(self):
        return cStringIO.StringIO(self.read())
        

def generate_files():
    """
    Generates all the Paths of interest to us.
    """

    dollspath = os.path.join(os.path.dirname(sys.argv[0]), "dolls")

    for dirpath, dirnames, filenames in os.walk(dollspath):

        dirnames[:] = [ i for i in dirnames if not i.startswith(".") ]
        filenames[:] = [ i for i in filenames if not i.startswith(".") ]

        for fn in filenames:
            yield Path(dirpath + '/' + fn)

            if fn.endswith(".zip") or fn.endswith(".azip"):
                zf = zipfile.ZipFile(dirpath + "/" + fn, "r")
                for i in zf.namelist():
                    yield ZipPath(zf, i)


def find_doll(doll):

    if doll not in dolls:
        dolls[doll] = Doll(doll)

    return dolls[doll]

def find_layer(doll, category, layer):

    d = find_doll(doll)

    if category not in d.layers:
        d.layers[category] = { }

    if layer not in d.layers[category]:
        d.layers[category][layer] = Layer(d, category, layer)

    return d.layers[category][layer]

def load_dolls():

    global dolls
    dolls = { }

    for p in generate_files():
        fn = p.basename
        fn = fn.lower()

        m = re.match(r'([^_]+)_([^_]+)_([^-+_]+)(_([^-+.]+))?([-+][\d.]+).spng', fn)
        if m:
            doll = m.group(1)
            category = m.group(2)
            layer = m.group(3)
            image = m.group(5) or layer
            priority = float(m.group(6))

            l = find_layer(doll, category, layer)

            im = Image(l, image, priority, p)
            l.images.append(im)

            continue

        m = re.match(r'([^_]+)_([^_]+)_([^-+.]+).tpng', fn)
        if m:
            doll = m.group(1)
            category = m.group(2)
            layer = m.group(3)

            l = find_layer(doll, category, layer)
            l.thumb = p

            continue

        if fn.endswith(".config.py"):

            doll = fn[:-len(".config.py")]
            d = find_doll(doll)
            
            d.config["__builtins__"] = dict(alice_version=1.0)
            exec p.read() in d.config, d.config

            continue

    
    # From here on in, we're traversing the doll data structure and
    # eliminating things that don't make sense. These includes:
    #
    # Layers with no images or no thumbnails.
    # Categories with no layers.
    # Dolls with no categories.

    for doll in dolls.keys():

        d = dolls[doll]

        for cat in d.layers.keys():
            for layer in d.layers[cat].keys():

                l = d.layers[cat][layer]

                # No thumb or no images.
                if not l.thumb or not l.images:
                    del d.layers[cat][layer]

            if not d.layers[cat]:
                del d.layers[cat]

        if not d.layers:
            del dolls[doll]

        
            
